/* dirbrowser.c - 2000/04/28 */
/*
 *  EasyTAG - Viewer/Editor for MP3 Tags
 *  Copyright (C) 2000  Jerome Couderc <j.couderc@ifrance.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/* Some parts of this browser are taken from:
 * XMMS - Cross-platform multimedia player
 * Copyright (C) 1998-1999  Peter Alm, Mikael Alm, Olle Hallnas,
 * Thomas Nilsson and 4Front Technologies
 */

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <sys/stat.h>
#include <dirent.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include "gdk/gdk.h"

#include "dirbrowser.h"
#include "easytag.h"
#include "msgbox.h"
#include "bar.h"
#include "misc.h"
#include "config.h"
#include "mylocale.h"

/* Pixmaps */
#include "../pixmaps/opened_folder.xpm"
#include "../pixmaps/closed_folder.xpm"
#include "../pixmaps/closed_folder_locked.xpm"

/* Returns the GList item for the nth row (taken from gtkclist.c) */
#define    ROW_ELEMENT(clist, row) (((row) == (clist)->rows - 1) ? \
                                   (clist)->row_list_end : \
                                   g_list_nth ((clist)->row_list, (row)))

/****************
 * Declarations *
 ****************/
static GdkPixmap *opened_folder_pixmap, *closed_folder_pixmap, *closed_folder_locked_pixmap;
static GdkBitmap *opened_folder_mask,   *closed_folder_mask,   *closed_folder_locked_mask;

GtkWidget *BrowserTree;
GtkWidget *BrowserList;
GtkCTreeNode *RootNode, *node, *nextnode;
GtkWidget *RenameDirectoryWindow = NULL;



/**************
 * Prototypes *
 **************/

static gint Browser_Tree_Key_Press (GtkWidget *ctree, GdkEvent *event);
void        Browser_Tree_Set_Node_Visible (GtkCTree *ctree, GtkCTreeNode *node);
void        Browser_Tree_Disable (void);
void        Browser_Tree_Enable (void);
void        Browser_Tree_Initialize (void);
void        Browser_Tree_Node_Selected (GtkCTree *tree, GtkCTreeNode *node, gint column, gpointer data);

void        Browser_List_Set_Row_Visible (GtkCList *ctree, gint row);
void        Browser_List_Disable (void);
void        Browser_List_Enable (void);

void        Browser_Entry_Disable (void);
void        Browser_Entry_Enable (void);

static gboolean check_for_subdir(gchar *path);
static void destroy_ctb (gpointer data);
static void destroy_clb (gpointer data);

gboolean    Check_For_Access_Permission (gchar *path);
static void expand_cb (GtkWidget *widget, GtkCTreeNode *parent_node);
static int  filetreeent_compare_func(const void *a, const void *b);

/* Pop up menus */
gboolean    Browser_Popup_Menu (GtkMenu *menu, GdkEventButton *event);
GtkWidget  *Create_Browser_Tree_Popup_Menu (GtkCTree *ctree);
GtkWidget  *Create_Browser_List_Popup_Menu (GtkCList *clist);

/* For window to rename a directory */
void        Open_Rename_Directory_Window (void);
void        Destroy_Rename_Directory_Window (void);
void        Rename_Directory (GtkObject *combobox);
void        Rename_Directory_Window_Key_Press (GtkWidget *window, GdkEvent *event);


/*************
 * Functions *
 *************/
/*
 * When you press enter in the Path entry to accept the text (browse the directory)
 */
void Path_Entry_Activate (void)
{
	gchar *path;

    Add_To_Combo_Box_History(GTK_OBJECT(BrowserEntry));
    path = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(BrowserEntry)->entry));
    Browser_Tree_Select_Dir(path);
}

/*
 * Set a text into the BrowserEntry (and don't activate it)
 */
void Browser_Entry_Set_Text (gchar *text)
{
	if (text && BrowserEntry)
	{
		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(BrowserEntry)->entry),text);
		/* Justify to the rigth, text into BrowserEntry */
		gtk_editable_set_position(GTK_EDITABLE(GTK_COMBO(BrowserEntry)->entry),-1);
	}
}


/*
 * Key Press events into browser
 */
static gint Browser_Tree_Key_Press (GtkWidget *ctree, GdkEvent *event)
{
	GdkEventKey *kevent;
	GList *SelectedNode;

	if (event && event->type==GDK_KEY_PRESS)
	{
		kevent = (GdkEventKey *)event;
		SelectedNode = ((GtkCList*)ctree)->selection;
		if (!SelectedNode) 
			return FALSE;
		switch(kevent->keyval)
		{
			case GDK_KP_Enter:    /* Enter key in Num Pad */
			case GDK_Return:      /* 'Normal' Enter key */
			case GDK_e:           /* Expand/Collapse node */
				gtk_ctree_toggle_expansion(GTK_CTREE(ctree),SelectedNode->data);
				return TRUE;
				break;

			case GDK_c:           /* Collapse node */
				gtk_ctree_collapse(GTK_CTREE(ctree),SelectedNode->data);
				return TRUE;
				break;
		}
	}
	return FALSE;
}


/*
 * Reload home directory
 */
void Load_Home_Dir (void)
{
	Browser_Tree_Select_Dir(HOME_VARIABLE);
}


/*
 * Reload default directory
 */
void Load_Default_Dir (void)
{
	Browser_Tree_Select_Dir(DEFAULT_PATH_TO_MP3);
}


/*
 * Get the path from the selected node (row) in the browser
 */
gchar *Get_Current_Path (void)
{
	GList *SelectedNode = NULL;
	DirNode *dirnode = NULL;

	SelectedNode = ((GtkCList*)BrowserTree)->selection;
	if (!SelectedNode) 
		return NULL;
	dirnode = gtk_ctree_node_get_row_data(GTK_CTREE(BrowserTree),SelectedNode->data);
	if (!dirnode) 
		return NULL;

    return dirnode->path;
}


/*
 * Set the current path in browser as default path
 */
void Set_Current_Path_As_Default (void)
{
    if (DEFAULT_PATH_TO_MP3 != NULL)
        g_free(DEFAULT_PATH_TO_MP3);
    DEFAULT_PATH_TO_MP3 = g_strdup(Get_Current_Path());
    Statusbar_Message(_("New default path for MP3 files selected"),TRUE);
}


/*
 * Collapse (close) tree recursively up to the root node.
 */
void Browser_Tree_Collapse (void)
{
	gtk_ctree_collapse_recursive(GTK_CTREE(BrowserTree),RootNode);
	/* But keep the main directory opened */
	gtk_ctree_expand(GTK_CTREE(BrowserTree),RootNode);
}


void Browser_Tree_Set_Node_Visible (GtkCTree *ctree, GtkCTreeNode *node)
{
	GtkVisibility visible;

	visible = gtk_ctree_node_is_visible(ctree,node);
	switch (visible)
	{
		case GTK_VISIBILITY_NONE:
		case GTK_VISIBILITY_PARTIAL:
			gtk_ctree_node_moveto(ctree,node,0,0.1,0.0);
			break;
		case GTK_VISIBILITY_FULL:
			break;
	}
}


void Browser_List_Set_Row_Visible (GtkCList *clist, gint row)
{
	GtkVisibility visible;

	visible = gtk_clist_row_is_visible(clist,row);
	switch (visible)
	{
		case GTK_VISIBILITY_NONE:
		case GTK_VISIBILITY_PARTIAL:
			gtk_clist_moveto(clist,row,0,0.5,0.0);
			break;
		case GTK_VISIBILITY_FULL:
			break;
	}
}


/*void Browser_Tree_Node_Selected (GtkCTree *tree, GList *node, gint column, gpointer data)*/
void Browser_Tree_Node_Selected (GtkCTree *tree, GtkCTreeNode *node, gint column, gpointer data)
{
	DirNode *dirnode;
	static int counter = 0;
/*	GList *SelectedDirectory = NULL;
 *	GtkCTreeNode *node;
 */
	/* Don't start a new reading, if an other one is running...*/
	if (ReadingDirectory == TRUE) return;

	Browser_Tree_Set_Node_Visible(tree,node);
	dirnode = gtk_ctree_node_get_row_data(GTK_CTREE(tree),node);
	if (dirnode->path)
	{
		/* Check if all files have been saved before to change directory */
		if (Check_If_All_Files_Are_Saved()==0)
		{
			GtkWidget *msgbox = NULL;
			gint button;

			msgbox = msg_box_new (_("Confirm..."),_("Some files have been modified but not "
				"saved...\nDo you want to save them before to change the directory?"),
				MSG_QUESTION,BUTTON_YES,BUTTON_NO,BUTTON_CANCEL,0);
			msg_box_hide_check_button(MSG_BOX(msgbox));
			button = msg_box_run(MSG_BOX(msgbox));
			switch(button)
			{
				case BUTTON_YES:
					if (Action_Save_All_Files_With_Answer()==-1) return;
					break;
				case BUTTON_NO:
					break;
				case BUTTON_CANCEL:
				case -1:
					return;
			}
		}
		/* Display the selected path into the BrowserEntry */
		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(BrowserEntry)->entry),dirnode->path);

		/* Start to read the directory */
		/* The first time 'counter' is equal to zero and if we don't want to load 
		 * directory on startup, we skip the 'reading', but we must read it */
		if (LOAD_ON_STARTUP || counter)
			Read_Directory(dirnode->path);
        else
            /* As we don't use the function 'Read_Directory' we must add this function here */
            Update_Command_Buttons_Sensivity();
		counter++;
	}

/************************************************************/
/* TESTING PART for selecting multiples lines in dirbrowser */
/************************************************************/
/*
	g_print("Entering Node_Selected..............\n");

	SelectedDirectory = GTK_CLIST(BrowserTree)->selection;
	g_print(">>>>'%d'\n",g_list_length(SelectedDirectory));
	SelectedDirectory = g_list_first(SelectedDirectory);
	while (SelectedDirectory)
	{
		gchar *tmp = g_strdup(SelectedDirectory->data);
		g_print(">'%x'\n",tmp);
		node = SelectedDirectory->data;
		dirnode = gtk_ctree_node_get_row_data(GTK_CTREE(BrowserTree),node);

		g_print(">'%s'\n",dirnode->path);

		if (SelectedDirectory->next) break;
		SelectedDirectory = g_list_next(SelectedDirectory);
	}
*/
}



/*
 * Browser_Tree_Select_Dir: Select the directory corresponding to the 'path' in
 * the tree browser, but it doesn't read it!
 * Check if path is correct before to select it. And returns 1 on success, else 0.
 */
gint Browser_Tree_Select_Dir (gchar *current_path)
{
	struct stat stbuf;
	gchar *currentdir,*pos,*tpath,*tpathnew;
	gboolean leaf;

    /* Inform the user if the full path is invalid, but load the first directories 
     * of the full path while it's valid */
	if ( stat(current_path,&stbuf)==-1)
	{
		GtkWidget *msgbox;
		gchar *msg;

		msg = g_strdup_printf(_("The entered path is invalid!:\n%s\n(%s)"),
		    current_path,g_strerror(errno));
		msgbox = msg_box_new (_("Error..."),msg,MSG_ERROR,BUTTON_OK,0);
		g_free(msg);
		msg_box_hide_check_button(MSG_BOX(msgbox));
		msg_box_run(MSG_BOX(msgbox));
		gtk_widget_destroy(msgbox);
		/*return 0;*/
	}

//	gtk_clist_freeze(GTK_CLIST(BrowserTree));
	/* Load current path */
	if(current_path && *current_path)
	{
		currentdir=g_strdup(current_path);
		tpath=g_strdup("/");
		gtk_ctree_expand(GTK_CTREE(BrowserTree),RootNode);
		pos=(gchar*)strtok(currentdir,"/");
		node=gtk_ctree_find_by_row_data_custom(GTK_CTREE(BrowserTree),NULL,"/",filetreeent_compare_func);
		do
		{
			tpathnew=g_strconcat(tpath,pos,"/",NULL);
			g_free(tpath);
			tpath=tpathnew;
			nextnode=gtk_ctree_find_by_row_data_custom(GTK_CTREE(BrowserTree),node,tpath,
								   filetreeent_compare_func);
			if(!nextnode)
				break;
			node=nextnode;
			pos=(gchar*)strtok(NULL,"/");
			gtk_ctree_get_node_info(GTK_CTREE(BrowserTree),node,NULL,NULL,NULL,NULL,NULL,NULL,&leaf,NULL);
			Browser_Tree_Set_Node_Visible(GTK_CTREE(BrowserTree),node);
			if(!leaf && pos)
				gtk_ctree_expand(GTK_CTREE(BrowserTree),node);
			else
			{
				gtk_ctree_select(GTK_CTREE(BrowserTree),node);
				break;
			}
		}
		while(pos);
		g_free(tpath);
		g_free(currentdir);
	}else
	{
		gtk_ctree_expand(GTK_CTREE(BrowserTree),RootNode);
	}
//	gtk_clist_thaw(GTK_CLIST(BrowserTree));
	return 1;
}


void Clist_Row_Selected (GtkCList *clist, gint row, gint column, GdkEventButton *event, gpointer data)
{
    FileRow *filerow = NULL;
    
    filerow = gtk_clist_get_row_data(GTK_CLIST(clist),row);
    Action_Select_Nth_File_By_Name(filerow->path);
}



/*
 * Load the list of mp3 files into the browser clist.
 */
void Browser_List_Load_Files (void)
{
	GdkColor bg_color = {0, 0xddd1, 0xeeec, 0xffff}; // Light blue
	gboolean activate_bg_color = 0;

	if (Mp3FileList && BrowserList)
	{
	    gtk_clist_freeze(GTK_CLIST(BrowserList));
        gtk_clist_clear(GTK_CLIST(BrowserList));
		Mp3FileList = g_list_first(Mp3FileList);
		while (Mp3FileList)
		{
            FileRow *filerow;
		    gchar *row_text[] = {NULL};
		    gint row;
		    
		    row_text[0] = g_strdup(g_basename( ((MP3_File *)Mp3FileList->data)->file) );

            row = gtk_clist_append(GTK_CLIST(BrowserList),row_text);
            /* Attached data */
        	filerow = g_malloc0(sizeof(FileRow));
        	filerow->path = g_strdup( ((MP3_File *)Mp3FileList->data)->file );
            gtk_clist_set_row_data_full(GTK_CLIST(BrowserList),row,filerow,destroy_clb);
            /* Toggle background color when directory is different */
            if (row != 0)
            {
                gchar *dir1;
                gchar *dir2;
                if (strcmp(dir1=g_dirname(((MP3_File *)Mp3FileList->prev->data)->file),
                           dir2=g_dirname(((MP3_File *)Mp3FileList->data)->file))!=0 )
                {
                    activate_bg_color = !activate_bg_color;
                }
                if (dir1) g_free(dir1);
                if (dir2) g_free(dir2);
            }
            if (activate_bg_color == TRUE)
                gtk_clist_set_background(GTK_CLIST(BrowserList),row,&bg_color);

		    if (!Mp3FileList->next) break;
		    Mp3FileList = Mp3FileList->next;
		}
		Mp3FileList = g_list_first(Mp3FileList);
	    gtk_clist_thaw(GTK_CLIST(BrowserList));
	}
}


void Browser_List_Select_File (gint row)
{
    if (row > 0)
    {
	    gtk_signal_handler_block_by_func(GTK_OBJECT(BrowserList),GTK_SIGNAL_FUNC(Clist_Row_Selected),NULL);
        gtk_clist_select_row(GTK_CLIST(BrowserList),row-1,0);
        Browser_List_Set_Row_Visible(GTK_CLIST(BrowserList),row-1);
	    gtk_signal_handler_unblock_by_func(GTK_OBJECT(BrowserList),GTK_SIGNAL_FUNC(Clist_Row_Selected),NULL);
    }
}


void Browser_List_Scroll_Vertical(GtkCList *clist, GtkScrollType scroll_type,
                                  gfloat position, gpointer user_data)
{
    if (!clist) return;
    
    /* As this function was called by signal gtk_signal_..._after..., "clist->focus_row" contains
     * the number of the row which receive the signal */ 
    gtk_clist_select_row(GTK_CLIST(clist),(gint)clist->focus_row,0);
}


void Browser_List_Clear (void)
{
    gtk_clist_clear(GTK_CLIST(BrowserList));
}


/*
 * Disable (FALSE) / Enable (TRUE) all user widgets in the browser area
 */
void Browser_Area_Set_Sensitive (gboolean activate)
{
    gtk_widget_set_sensitive(GTK_WIDGET(BrowserEntry),activate);
    gtk_widget_set_sensitive(GTK_WIDGET(BrowserTree), activate);
    gtk_widget_set_sensitive(GTK_WIDGET(BrowserList), activate);
}


/*gint Browser_Tree_Width (void)
{
    return BrowserTree->parent->allocation.width;
}
gint Browser_Tree_Height (void)
{
    return BrowserTree->parent->allocation.height;
}
*/


/*
 * Create_Browser_Tree_Popup_Menu: Create a popup menu for the tree browser
 * Create_Browser_List_Popup_Menu: Create a popup menu for the list of files of browser
 */
gboolean Browser_Popup_Menu_Handler (GtkMenu *menu, GdkEventButton *event)
{
	if (event && (event->type==GDK_BUTTON_PRESS) && (event->button==3))
	{
		gtk_menu_popup(menu,NULL,NULL,NULL,NULL,event->button,event->time);
		return TRUE;
	}
	return FALSE;
}
GtkWidget *Create_Browser_Tree_Popup_Menu (GtkCTree *ctree)
{
	GtkWidget *BrowserPopupMenu;
	GtkWidget *MenuItem;


	BrowserPopupMenu = gtk_menu_new();
	gtk_signal_connect_object(GTK_OBJECT(BrowserTree),"button_press_event",
				  (GtkSignalFunc)Browser_Popup_Menu_Handler,
				  GTK_OBJECT(BrowserPopupMenu));

	MenuItem = gtk_menu_item_new_with_label(_("Go to Home Directory"));
	gtk_menu_append(GTK_MENU(BrowserPopupMenu),MenuItem);
	gtk_signal_connect(GTK_OBJECT(MenuItem),"activate",(GtkSignalFunc)Load_Home_Dir,NULL);

	MenuItem = gtk_menu_item_new_with_label(_("Go to Default Directory"));
	gtk_menu_append(GTK_MENU(BrowserPopupMenu),MenuItem);
	gtk_signal_connect(GTK_OBJECT(MenuItem),"activate",(GtkSignalFunc)Load_Default_Dir,NULL);

	MenuItem = gtk_menu_item_new_with_label(_("Set Current Path as Default"));
	gtk_menu_append(GTK_MENU(BrowserPopupMenu),MenuItem);
	gtk_signal_connect(GTK_OBJECT(MenuItem),"activate",(GtkSignalFunc)Set_Current_Path_As_Default,NULL);

	MenuItem = gtk_menu_item_new();
	gtk_menu_append(GTK_MENU(BrowserPopupMenu),MenuItem);

	MenuItem = gtk_menu_item_new_with_label(_("Rename Directory..."));
	gtk_menu_append(GTK_MENU(BrowserPopupMenu),MenuItem);
	gtk_signal_connect(GTK_OBJECT(MenuItem),"activate",(GtkSignalFunc)Open_Rename_Directory_Window,NULL);

	MenuItem = gtk_menu_item_new();
	gtk_menu_append(GTK_MENU(BrowserPopupMenu),MenuItem);

	CheckMenuItemBrowseSubdirPopupMenu = gtk_check_menu_item_new_with_label(_("Browse Sub-directories"));
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(CheckMenuItemBrowseSubdirPopupMenu),BROWSE_SUBDIR);
	gtk_check_menu_item_set_show_toggle(GTK_CHECK_MENU_ITEM(CheckMenuItemBrowseSubdirPopupMenu),TRUE);
	gtk_menu_append(GTK_MENU(BrowserPopupMenu),CheckMenuItemBrowseSubdirPopupMenu);
	gtk_signal_connect(GTK_OBJECT(CheckMenuItemBrowseSubdirPopupMenu),"toggled",
		(GtkSignalFunc)Check_Menu_Item_Toggled_Browse_Subdir,NULL);

	MenuItem = gtk_menu_item_new();
	gtk_menu_append(GTK_MENU(BrowserPopupMenu),MenuItem);

	MenuItem = gtk_menu_item_new_with_label(_("Collapse Tree"));
	gtk_menu_append(GTK_MENU(BrowserPopupMenu),MenuItem);
	gtk_signal_connect(GTK_OBJECT(MenuItem),"activate",(GtkSignalFunc)Browser_Tree_Collapse,NULL);

	MenuItem = gtk_menu_item_new_with_label(_("Refresh Tree"));
	gtk_menu_append(GTK_MENU(BrowserPopupMenu),MenuItem);
	gtk_signal_connect(GTK_OBJECT(MenuItem),"activate",(GtkSignalFunc)Browser_Tree_Rebuild,NULL);

	MenuItem = gtk_menu_item_new();
	gtk_menu_append(GTK_MENU(BrowserPopupMenu),MenuItem);

	MenuItem = gtk_menu_item_new_with_label(_("Run XMMS"));
	gtk_menu_append(GTK_MENU(BrowserPopupMenu),MenuItem);
	gtk_signal_connect(GTK_OBJECT(MenuItem),"activate",(GtkSignalFunc)Run_Xmms,NULL);

	gtk_widget_show_all(BrowserPopupMenu);
	return BrowserPopupMenu;
}
GtkWidget  *Create_Browser_List_Popup_Menu (GtkCList *clist)
{
	GtkWidget *BrowserPopupMenu;
	GtkWidget *MenuItem;


	BrowserPopupMenu = gtk_menu_new();
	gtk_signal_connect_object(GTK_OBJECT(BrowserList),"button_press_event",
				  (GtkSignalFunc)Browser_Popup_Menu_Handler,
				  GTK_OBJECT(BrowserPopupMenu));

	MenuItem = gtk_menu_item_new_with_label(_("Sort list ascending..."));
	gtk_menu_append(GTK_MENU(BrowserPopupMenu),MenuItem);
	gtk_signal_connect(GTK_OBJECT(MenuItem),"activate",(GtkSignalFunc)Sort_Mp3list_Ascending,NULL);
    
	MenuItem = gtk_menu_item_new_with_label(_("Sort list descending..."));
	gtk_menu_append(GTK_MENU(BrowserPopupMenu),MenuItem);
	gtk_signal_connect(GTK_OBJECT(MenuItem),"activate",(GtkSignalFunc)Sort_Mp3list_Descending,NULL);
    
	gtk_widget_show_all(BrowserPopupMenu);
	return BrowserPopupMenu;
}


gint Find_Node (gconstpointer a, gconstpointer b)
{
	gchar *data1;
	gchar *data2;

	if (!a || !b || !((DirNode*)a)->path) return -1;
	data1 = g_strdup((gchar *)((DirNode*)a)->path);
	data2 = g_strdup((gchar *)b);

	if ( strncmp(data1,data2,strlen(data2))==0 && strlen(data1)>strlen(data2) )
	{
		g_free(data1);
		g_free(data2);
		return 0;
	}else
	{
		g_free(data1);
		g_free(data2);
		return -1;
	}
}

/*
 * Destroy the whole tree up to the root node
 */
void Browser_Tree_Initialize (void)
{
	GtkCTreeNode *node;
	DirNode *dirnode;
	gchar *node_text = "dummy";

	gtk_clist_freeze(GTK_CLIST(BrowserTree));
	dirnode = gtk_ctree_node_get_row_data(GTK_CTREE(BrowserTree),RootNode);
	while ( (node=gtk_ctree_find_by_row_data_custom(GTK_CTREE(BrowserTree),NULL,dirnode->path,Find_Node)) )
	{
		gtk_ctree_remove_node(GTK_CTREE(BrowserTree),node);
	}
	/* Reinit node data */
	dirnode->scanned = 0;
	/* Insert dummy node to display: '+' */
	node = gtk_ctree_insert_node(GTK_CTREE(BrowserTree),RootNode,NULL,&node_text,4,
		NULL,NULL,NULL,NULL,TRUE,TRUE);
	gtk_clist_thaw(GTK_CLIST(BrowserTree));
}

/*
 * Browser_Tree_Rebuild: Refresh the tree browser by destroying it and rebuilding it.
 * Opens tree nodes corresponding to 'path_to_load' if this parameter isn't NULL.
 * If NULL, selects the current path.
 */
void Browser_Tree_Rebuild (gchar *path_to_load)
{
    if (path_to_load != NULL)
    {
        Browser_Tree_Initialize();
    	Browser_Tree_Select_Dir(path_to_load);
    }else
    {
    	gchar *current_path = NULL;

    	/* Memorize the current path to load it again at the end */
    	current_path = g_strdup(Get_Current_Path());
    	/* If no node selected, get path from BrowserEntry or default path */
    	if (current_path == NULL)
    	{
    		if (strlen(gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(BrowserEntry)->entry)))>0)
    			current_path = g_strdup(gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(BrowserEntry)->entry)));
    		else
    			current_path = g_strdup(DEFAULT_PATH_TO_MP3);
    	}
    	
        Browser_Tree_Initialize();
    	/* Reload the memorized path */
//        gtk_signal_handler_block_by_func(GTK_OBJECT(BrowserTree),GTK_SIGNAL_FUNC(Browser_Tree_Node_Selected),NULL);
    	Browser_Tree_Select_Dir(current_path);
//        gtk_signal_handler_unblock_by_func(GTK_OBJECT(BrowserTree),GTK_SIGNAL_FUNC(Browser_Tree_Node_Selected),NULL);
    	if (current_path) g_free(current_path);
    }
}



static gboolean check_for_subdir(gchar *path)
{
	DIR *dir;
	struct dirent *dirent;
	struct stat statbuf;
	gchar *npath;
	
	if( (dir=opendir(path)) )
	{
		while( (dirent=readdir(dir)) )
		{
			if(dirent->d_name[0]!='.')
			{
				npath = g_strconcat(path,dirent->d_name,"/",NULL);
				stat(npath,&statbuf);
				g_free(npath);
				if(S_ISDIR(statbuf.st_mode))
				{
					closedir(dir);
					return TRUE;
				}
			}
		}
		closedir(dir);
	}
	return FALSE;
}
	
static void destroy_ctb(gpointer data)
{
	DirNode *node = data;
	g_free(node->path);
	g_free(node);
}
static void destroy_clb(gpointer data)
{
	FileRow *row = data;
	g_free(row->path);
	g_free(row);
}


/*
 * Check if you have access permissions for directory path. Returns 1 if ok, else 0.
 */
gboolean Check_For_Access_Permission (gchar *path)
{
	DIR *dir;

	if( (dir=opendir(path)) == NULL )
	{
		if (errno == EACCES)
			return FALSE;
	}else
	{
		closedir(dir);
	}
	return TRUE;
}


static void expand_cb(GtkWidget *ctree, GtkCTreeNode *parent_node)
{
	DIR *dir;
	struct dirent *dirent;
	gchar *path;
	gchar *text;
	gchar *dummy = "dummy";
	struct stat statbuf;
	GtkCTreeNode *node, *sub_node;
	DirNode *parent_dirnode, *dirnode;
	gboolean has_subdir = FALSE;
	
	
	parent_dirnode = gtk_ctree_node_get_row_data(GTK_CTREE(ctree), parent_node);
	if(!parent_dirnode->scanned)
	{
		gtk_clist_freeze(GTK_CLIST(ctree));
		node = gtk_ctree_find_by_row_data(GTK_CTREE(ctree), parent_node, NULL);
		gtk_ctree_remove_node(GTK_CTREE(ctree),node);
		if( (dir=opendir(parent_dirnode->path)) )
		{
			while( (dirent=readdir(dir)) )
			{
				path = g_strconcat(parent_dirnode->path,dirent->d_name,NULL);
				stat(path,&statbuf);
				if(S_ISDIR(statbuf.st_mode)&&dirent->d_name[0]!='.')
				{
					dirnode = g_malloc0(sizeof(DirNode));
					dirnode->path = g_strconcat(path,"/",NULL);
					text = dirent->d_name;

					if(check_for_subdir(dirnode->path))
						has_subdir=TRUE;
					else
						has_subdir=FALSE;

					/* Select pixmap for accessible/unaccessible directory */
					if (Check_For_Access_Permission(path))
					{
						node = gtk_ctree_insert_node(GTK_CTREE(ctree),
							parent_node,NULL,&text,4, 
							closed_folder_pixmap,closed_folder_mask,
							opened_folder_pixmap,opened_folder_mask, 
							!has_subdir, FALSE);
					}
					else
					{
						node = gtk_ctree_insert_node(GTK_CTREE(ctree),
							parent_node,NULL,&text,4, 
							closed_folder_locked_pixmap,closed_folder_locked_mask,
							opened_folder_pixmap,opened_folder_mask, 
							!has_subdir, FALSE);
					}

					gtk_ctree_node_set_row_data_full(GTK_CTREE(ctree),node,dirnode,destroy_ctb);

					if(has_subdir)
						sub_node=gtk_ctree_insert_node(GTK_CTREE(ctree),
							node,NULL,&dummy,4,NULL,NULL,NULL,NULL,FALSE,FALSE);
				}
				g_free(path);
			}
			closedir(dir);
			gtk_ctree_sort_node(GTK_CTREE(ctree),parent_node);
		}
		gtk_clist_thaw(GTK_CLIST(ctree));	
		parent_dirnode->scanned=TRUE;
	}
}

static int filetreeent_compare_func(const void *a, const void *b)
{
	if(!a || !b || !((DirNode*)a)->path)
		return -1;
	return strcmp(((DirNode*)a)->path,(gchar*)b);
}


/*
 * Create the browser
 */
GtkWidget *Create_Browser_Tree (GtkWidget *parent)
{
    GtkWidget *VerticalBox;
    GtkWidget *ScrollWindowCTree;
    GtkWidget *ScrollWindowCList;
    gchar *RootText = "/";
    gchar *node_text = "dummy";
    DirNode *dirnode;
    GtkTooltips *Tips;
    GtkWidget *PopupMenu;
    GList *History_List;


	Tips = gtk_tooltips_new();
	Set_Tips_Color(Tips);

	VerticalBox = gtk_vbox_new(FALSE,2);
	gtk_container_set_border_width(GTK_CONTAINER(VerticalBox),2);

	/*
	 * The entry box for displaying path
	 */
	BrowserEntry = gtk_combo_new();
	gtk_combo_disable_activate(GTK_COMBO(BrowserEntry));
    /* History list */
    History_List = Load_Path_Entry_List();
    if (History_List)
        gtk_combo_set_popdown_strings(GTK_COMBO(BrowserEntry),History_List);
    gtk_object_set_data(GTK_OBJECT(BrowserEntry),"History",History_List);

	gtk_signal_connect(GTK_OBJECT(GTK_COMBO(BrowserEntry)->entry),"activate",GTK_SIGNAL_FUNC(Path_Entry_Activate),NULL);
	gtk_box_pack_start(GTK_BOX(VerticalBox),BrowserEntry,FALSE,TRUE,0);
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(BrowserEntry)->entry),_("No path selected!"));
	gtk_tooltips_set_tip(Tips,GTK_COMBO(BrowserEntry)->entry,_("Enter a directory to browse."),NULL);

	/*
	 * The ScrollWindow and the CTree
	 */
	ScrollWindowCTree = gtk_scrolled_window_new(NULL,NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(ScrollWindowCTree),
	                               GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC);
	//gtk_box_pack_start(GTK_BOX(VerticalBox),ScrollWindowCTree,TRUE,TRUE,0);
	///gtk_widget_set_usize (GTK_WIDGET(ScrollWindow),150,100);

	/* The tree */
	BrowserTree = gtk_ctree_new(1,0);
	gtk_container_add(GTK_CONTAINER(ScrollWindowCTree),BrowserTree);

 	/* Tree config */
	gtk_ctree_set_indent(GTK_CTREE(BrowserTree),13);
    gtk_ctree_set_line_style(GTK_CTREE(BrowserTree),GTK_CTREE_LINES_DOTTED);
	gtk_ctree_set_expander_style(GTK_CTREE(BrowserTree),GTK_CTREE_EXPANDER_SQUARE);
	gtk_clist_set_column_auto_resize(GTK_CLIST(BrowserTree),0,TRUE);
    /*gtk_clist_set_selection_mode (GTK_CLIST(BrowserTree),GTK_SELECTION_EXTENDED);*/

	/* Create pixmaps */
	if(!opened_folder_pixmap)
	{
		opened_folder_pixmap = gdk_pixmap_create_from_xpm_d(parent->window,
					&opened_folder_mask,NULL,opened_folder_xpm);
		closed_folder_pixmap = gdk_pixmap_create_from_xpm_d(parent->window,
					&closed_folder_mask,NULL,closed_folder_xpm);
		closed_folder_locked_pixmap = gdk_pixmap_create_from_xpm_d(parent->window,
					&closed_folder_locked_mask,NULL,closed_folder_locked_xpm);
	}
	
	/* Signaux */
	gtk_signal_connect(GTK_OBJECT(BrowserTree),"tree_expand",    (GtkSignalFunc)expand_cb,NULL);
	gtk_signal_connect(GTK_OBJECT(BrowserTree),"tree_select_row",(GtkSignalFunc)Browser_Tree_Node_Selected,NULL);
	gtk_signal_connect(GTK_OBJECT(BrowserTree),"key_press_event",(GtkSignalFunc)Browser_Tree_Key_Press,NULL);
	gtk_signal_connect_after(GTK_OBJECT(BrowserTree),"scroll_vertical",(GtkSignalFunc)Browser_List_Scroll_Vertical,NULL);

	/* Create the root node */
	RootNode = gtk_ctree_insert_node(GTK_CTREE(BrowserTree),NULL,NULL,&RootText,4, 
					closed_folder_pixmap,closed_folder_mask,
					opened_folder_pixmap,opened_folder_mask, 
					FALSE, FALSE);

	dirnode = g_malloc0(sizeof(DirNode));
	dirnode->path = g_strdup("/");
	gtk_ctree_node_set_row_data_full(GTK_CTREE(BrowserTree),RootNode,dirnode,destroy_ctb);

	/* Insert dummy node to display: '+' */
	node = gtk_ctree_insert_node(GTK_CTREE(BrowserTree),RootNode,NULL,&node_text,4, 
					NULL,NULL,NULL,NULL,TRUE,TRUE);

	/* 
	 * Create Popup Menu on browser ctree
	 */
	PopupMenu = Create_Browser_Tree_Popup_Menu(GTK_CTREE(BrowserTree));

	/* Open root node */
	/*gtk_ctree_expand(GTK_CTREE(BrowserTree),RootNode);*/

	/* Load current path */
	/* Don't load it now! */

	/*
	 * The ScrollWindow and the CList
	 */
	ScrollWindowCList = gtk_scrolled_window_new(NULL,NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(ScrollWindowCList),
	                               GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC);

	/* The clist */
	BrowserList = gtk_clist_new(1);
	gtk_container_add(GTK_CONTAINER(ScrollWindowCList),BrowserList);
	gtk_clist_set_column_auto_resize(GTK_CLIST(BrowserList),0,TRUE);
	gtk_clist_set_reorderable(GTK_CLIST(BrowserList),FALSE);
	gtk_clist_set_selection_mode(GTK_CLIST(BrowserList),GTK_SELECTION_SINGLE);

	gtk_signal_connect(GTK_OBJECT(BrowserList),"select_row",(GtkSignalFunc)Clist_Row_Selected,NULL);
    /* Note: "after" for this signal to obtain the new focus_row of clist and not the previous one */
	gtk_signal_connect_after(GTK_OBJECT(BrowserList),"scroll_vertical",
	    (GtkSignalFunc)Browser_List_Scroll_Vertical,NULL);

	/* 
	 * Create Popup Menu on browser ctree
	 */
	PopupMenu = Create_Browser_List_Popup_Menu(GTK_CLIST(BrowserList));


	/*
	 * The pane for the CTree and the CList
	 */
    BrowserHPaned = gtk_hpaned_new();
	gtk_box_pack_start(GTK_BOX(VerticalBox),BrowserHPaned,TRUE,TRUE,0);
    gtk_paned_set_handle_size(GTK_PANED(BrowserHPaned),8);
    gtk_paned_set_gutter_size(GTK_PANED(BrowserHPaned),6);                       
    gtk_paned_pack1(GTK_PANED(BrowserHPaned),ScrollWindowCTree,TRUE,FALSE);
    gtk_paned_pack2(GTK_PANED(BrowserHPaned),ScrollWindowCList,TRUE,FALSE);
    if (SET_PANE_HANDLE_POSITION2)
        gtk_paned_set_position(GTK_PANED(BrowserHPaned),PANE_HANDLE_POSITION2);

	gtk_widget_show_all(VerticalBox);
	return VerticalBox;
}



/*
 * The window to rename a directory into the browser.
 */
void Open_Rename_Directory_Window (void)
{
	GtkWidget *Frame;
	GtkWidget *VBox;
	GtkWidget *Combo;
	GtkWidget *ButtonBox;
	GtkWidget *Button;
	GtkWidget *Separator;
	GList *combo_list = NULL;
	gchar *directory_parent = NULL;
	gchar *directory_name = NULL;
	gchar *address = NULL;
	gint mw_x, mw_y;
	gint ba_x, ba_y;


    if (RenameDirectoryWindow != NULL) 
    {
        gdk_window_raise(RenameDirectoryWindow->window);
        return;
    }

    /* We get the full path but we musn't display the parent directories */
	directory_parent = g_strdup(Get_Current_Path());
    if (!directory_parent || strlen(directory_parent)==0)
        return;
	if (strlen(directory_parent)>1 && directory_parent[strlen(directory_parent)-1]=='/')
	    directory_parent[strlen(directory_parent)-1]=0;
	address = strrchr(directory_parent,'/');
	if (!address) return;
	directory_name = g_strdup(address+1);
	*(address+1) = 0;
    if (!directory_name || strlen(directory_name)==0)
        return;
    
    
	RenameDirectoryWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(RenameDirectoryWindow),_("Rename the directory"));
	gtk_window_set_transient_for(GTK_WINDOW(RenameDirectoryWindow),GTK_WINDOW(MainWindow));
	gtk_window_set_policy(GTK_WINDOW(RenameDirectoryWindow),FALSE,TRUE,TRUE);
	gtk_signal_connect(GTK_OBJECT(RenameDirectoryWindow),"destroy",
		(GtkSignalFunc)Destroy_Rename_Directory_Window,NULL);
	gtk_signal_connect(GTK_OBJECT(RenameDirectoryWindow),"delete_event",
		(GtkSignalFunc)Destroy_Rename_Directory_Window,NULL);
	gtk_signal_connect(GTK_OBJECT(RenameDirectoryWindow),"key_press_event",
	    (GtkSignalFunc)Rename_Directory_Window_Key_Press,NULL);

	Frame = gtk_frame_new(NULL);
	gtk_container_add(GTK_CONTAINER(RenameDirectoryWindow),Frame);
	gtk_container_set_border_width(GTK_CONTAINER(Frame),2);

	VBox = gtk_vbox_new(FALSE,0);
	gtk_container_add(GTK_CONTAINER(Frame),VBox);
	gtk_container_border_width(GTK_CONTAINER(VBox),4);

	/* The combobox to rename the directory */
	Combo = gtk_combo_new();
	gtk_container_border_width(GTK_CONTAINER(Combo),4);
	gtk_box_pack_start(GTK_BOX(VBox),Combo,TRUE,TRUE,4);
	gtk_combo_set_value_in_list(GTK_COMBO(Combo),FALSE,FALSE);
	gtk_widget_set_usize(GTK_WIDGET(Combo),300,-1);
	/* Set the directory into the combobox */
    combo_list = g_list_append(combo_list,"");
    combo_list = g_list_append(combo_list,directory_name);
//	gtk_combo_set_popdown_strings(GTK_COMBO(Combo),combo_list);
//*** FIX ME! When activated, can't change letter case
    gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(Combo)->entry),directory_name);
    /* We attach usefull data to the combobox */
    gtk_object_set_data(GTK_OBJECT(Combo),"Parent_Directory",directory_parent);
    gtk_object_set_data(GTK_OBJECT(Combo),"Current_Directory",directory_name);
    Attach_Popup_Menu_To_Tag_Entries(GTK_ENTRY(GTK_COMBO(Combo)->entry));

	/* Separator line */
	Separator = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(VBox),Separator,FALSE,FALSE,0);

	ButtonBox = gtk_hbutton_box_new ();
	gtk_box_pack_start(GTK_BOX(VBox),ButtonBox,FALSE,FALSE,0);
	gtk_button_box_set_layout(GTK_BUTTON_BOX(ButtonBox),GTK_BUTTONBOX_END);
	gtk_button_box_set_spacing(GTK_BUTTON_BOX(ButtonBox),10);

    /* Button to rename directory */
	Button = Create_Button_With_Pixmap(BUTTON_SAVE);
	gtk_container_add(GTK_CONTAINER(ButtonBox),Button);
	GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
	gtk_signal_connect_object(GTK_OBJECT(Button),"clicked",(GtkSignalFunc)Rename_Directory,GTK_OBJECT(Combo));
	gtk_signal_connect_object(GTK_OBJECT(GTK_COMBO(Combo)->entry),"changed",(GtkSignalFunc)Entry_Changed_Disable_Object,GTK_OBJECT(Button));
	
    /* Button to cancel */
	Button = Create_Button_With_Pixmap(BUTTON_CANCEL);
	gtk_container_add(GTK_CONTAINER(ButtonBox),Button);
	GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
	gtk_widget_grab_default (Button);
	gtk_signal_connect(GTK_OBJECT(Button),"clicked",(GtkSignalFunc)Destroy_Rename_Directory_Window,NULL);

	gtk_widget_show_all(RenameDirectoryWindow);

    /* Note : the window must be shown before to move it */
	gdk_window_get_position(MainWindow->window,&mw_x,&mw_y);
	gdk_window_get_position(BrowseArea->window,&ba_x,&ba_y);
    gtk_window_reposition(GTK_WINDOW(RenameDirectoryWindow),
		mw_x + ba_x + (BrowseArea->allocation.width/2)  - (RenameDirectoryWindow->allocation.width)/2,
		mw_y + ba_y + (BrowseArea->allocation.height/2) - (RenameDirectoryWindow->allocation.height)/2);
}
void Destroy_Rename_Directory_Window (void)
{
	if (RenameDirectoryWindow)
	{
		gtk_widget_destroy(RenameDirectoryWindow);
		RenameDirectoryWindow = (GtkWidget *)NULL;
	}
}
void Rename_Directory (GtkObject *combobox)
{
    gchar *directory_parent;
    gchar *directory_last_name;
    gchar *directory_new_name;
    gchar *last_path;
    gchar *tmp_path;
    gchar *new_path;
    struct stat statbuf;


    if (!GTK_IS_COMBO(combobox)) return;
    
    directory_parent    = gtk_object_get_data(GTK_OBJECT(combobox),"Parent_Directory");
    directory_last_name = gtk_object_get_data(GTK_OBJECT(combobox),"Current_Directory");
    directory_new_name  = gtk_editable_get_chars(GTK_EDITABLE(GTK_COMBO(combobox)->entry),0,-1);

    /* Check if a name for the directory have been supplied */
    if (!directory_new_name || strlen(directory_new_name)<1)
    {
		GtkWidget *msgbox;

		msgbox = msg_box_new (_("Error..."),_("You must type a directory name!"),MSG_ERROR,BUTTON_OK,0);
		msg_box_hide_check_button(MSG_BOX(msgbox));
		msg_box_run(MSG_BOX(msgbox));
		gtk_widget_destroy(msgbox);
		return;
    }

    /* If the directory name haven't been changed, we do nothing! */
    if (directory_last_name && directory_new_name && strcmp(directory_last_name,directory_new_name)==0)
    {
        Destroy_Rename_Directory_Window();
		return;
	}

    /* Build the current and new absolute paths */
    last_path = g_strconcat(directory_parent,directory_last_name,NULL);
    new_path  = g_strconcat(directory_parent,directory_new_name,NULL);
    /* Temporary path (useful when changing only string case) */
    tmp_path = g_strconcat(new_path,".EastTAG",NULL);


    /* Check if the new directory name doesn' already exists (case change is accepted) */
	stat(new_path,&statbuf);
	if(S_ISDIR(statbuf.st_mode) && strcmp(last_path,new_path)==0)
    {
		gchar *msg;
		GtkWidget *msgbox;

		msg = g_strdup_printf(_("Can't rename because this directory name "
		                        "already exists!\n(%s)"),new_path); 
		msgbox = msg_box_new (_("Error..."),msg,MSG_ERROR,BUTTON_OK,0);
		g_free(msg);
		msg_box_hide_check_button(MSG_BOX(msgbox));
		msg_box_run(MSG_BOX(msgbox));
		gtk_widget_destroy(msgbox);

		return;
    }

    /* Rename the directory from last name to tmp name */
	if ( rename(last_path,tmp_path)!=0 )
	{
		gchar *msg;
		GtkWidget *msgbox;

		msg = g_strdup_printf(_("Can't rename directory \n'%s'\n to \n'%s'!\n(%s)"),
					last_path,tmp_path,g_strerror(errno)); 
		msgbox = msg_box_new (_("Error..."),msg,MSG_ERROR,BUTTON_OK,0);
		g_free(msg);
		msg_box_hide_check_button(MSG_BOX(msgbox));
		msg_box_run(MSG_BOX(msgbox));
		gtk_widget_destroy(msgbox);

		return;
	}
    /* Rename the directory from tmp name to new name (final name) */
	if ( rename(tmp_path,new_path)!=0 )
	{
		gchar *msg;
		GtkWidget *msgbox;

		msg = g_strdup_printf(_("Can't rename directory \n'%s'\n to \n'%s'!\n(%s)"),
					tmp_path,new_path,g_strerror(errno)); 
		msgbox = msg_box_new (_("Error..."),msg,MSG_ERROR,BUTTON_OK,0);
		g_free(msg);
		msg_box_hide_check_button(MSG_BOX(msgbox));
		msg_box_run(MSG_BOX(msgbox));
		gtk_widget_destroy(msgbox);

		return;
	}

    g_free(tmp_path);

    Destroy_Rename_Directory_Window();
    Browser_Tree_Rebuild(new_path);
    Statusbar_Message(_("Directory renamed"),TRUE);
}
void Rename_Directory_Window_Key_Press (GtkWidget *window, GdkEvent *event)
{
	GdkEventKey *kevent;

	if (event && event->type == GDK_KEY_PRESS)
	{
		kevent = (GdkEventKey *)event;
		switch(kevent->keyval)
		{
			case GDK_Escape:
				Destroy_Rename_Directory_Window();
				break;
		}
    }
}
