/* header.c - 2000/05/12 */
/*
 *  EasyTAG - Viewer/Editor for MP3 Tags
 *  Copyright (C) 2000  Jerome Couderc <j.couderc@ifrance.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <gtk/gtk.h>
#include <stdio.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>

#include "header.h"
#include "easytag.h"
#include "mylocale.h"


/****************
 * Declarations *
 ****************/

/* Header Content Presentation:
 * For more informations, visit web page: http://www.dv.co.yu/mpgscript/mpeghdr.htm
 *
 *  31          25        20        15        10         5         0
 *  |7 6 5 4 3 2 1 0|7 6 5 4 3 2 1 0|7 6 5 4 3 2 1 0|7 6 5 4 3 2 1 0|
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |         SYNCH       |  V| L |E|  BRI  | SF|P|X| MO| MX|C|O| EM|
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *
 *  SYNCH        = frame synchronisation (all bits set to 1) : 11 bits
 *  V            = MPEG version (11=MPEG 1, 10=MPEG 2, 00-MPEG 2.5)
 *  L            = layer (11=I, 10=II, 01=III)
 *  E            = use error-protection: 0=CRC (16bits crc follows header), 1=not protected
 *  BRI          = bitrate index
 *  SF           = sample frequency
 *  P            = padding
 *  X            = extension
 *  MO           = mode (11=single_channel, 10=dual_channel, 01=joint_stereo, 00=stereo)
 *  MX           = mode extension
 *  C            = copyright
 *  O            = original
 *  EM           = emphasis   
 *
 */

char *version_names[3] =
{
	"1",    /* 11 - MPEG 1 */
	"2",    /* 10 - MPEG 2 */
	"2.5"   /* 00 - MPEG 2.5 */
};

char *layer_names[3] =
{
	"I",    /* 11 - Layer 1 */
	"II",   /* 10 - Layer 2 */
	"III"   /* 01 - Layer 3 */
};

char* mode_names[4] =
{
	N_("Stereo"),
	N_("Joint Stereo"),    /* Stereo */
	N_("Dual Channel"),    /* Stereo */
	N_("Single Channel")   /* Mono */
};

int frequencies[3][4] =
{
	{44100, 48000, 32000, 0},    /* MPEG 1 */
	{22050, 24000, 16000, 0},    /* MPEG 2 */
	{11025,  8000,  8000, 0}     /* MPEG 2.5 */
};

int bitrates[3][3][14] =
{
  {		/* MPEG 1 */
      {32,64,96,128,160,192,224,256,288,320,352,384,416,448},    /* Layer 1 */
      {32,48,56, 64, 80, 96,112,128,160,192,224,256,320,384},    /* Layer 2 */
      {32,40,48, 56, 64, 80, 96,112,128,160,192,224,256,320}     /* Layer 3 */
  },
  {		/* MPEG 2 */
      {32,48,56,64,80,96,112,128,144,160,176,192,224,256},    /* Layer 1 */
      { 8,16,24,32,40,48, 56, 64, 80, 96,112,128,144,160},    /* Layer 2 */
      { 8,16,24,32,40,48, 56, 64, 80, 96,112,128,144,160}     /* Layer 3 */
  },
  {		/* MPEG 2.5 */
      {32,48,56,64,80,96,112,128,144,160,176,192,224,256},    /* Layer 1 */
      { 8,16,24,32,40,48, 56, 64, 80, 96,112,128,144,160},    /* Layer 2 */
      { 8,16,24,32,40,48, 56, 64, 80, 96,112,128,144,160}     /* Layer 3 */
  }
};


/**************
 * Prototypes *
 **************/
glong Header_Find_Position (FILE *file, long start_position);
gchar *Convert_Size (gfloat size);
gchar *Convert_Duration (long int duration);


/*************
 * Functions *
 *************/
/*
 * Read infos of 1rst header frame
 */
void Header_Read_Mp3 (MP3_File *mp3file)
{
	FILE *file;
	gchar buffer[MPEG_HEADER_LENGTH];
	glong position = 0;
	Mpeg_Info mpegfile;
	struct stat statbuf;
	gint filesize;

	if ( (file=fopen(mp3file->file,"r"))==0 )
	{
		g_print(_("ERROR while opening file: '%s' (%s)\n\a"),mp3file->file,g_strerror(errno));
		return ;
	}

	/* For debugging */
	/*g_print("Opening file: '%s'\n",g_basename(mp3file->file));*/

	/* Loop while a valid header is not found, if end of file is reached 'return' is emitted */
	while (1)
	{
		position = Header_Find_Position(file,position);
		if (position == -1)
		{
			/* End of file reached */
			g_print("Can't find header in file: '%s'\n",g_basename(mp3file->file));
			/* Close the file */
			if ( !file || fclose(file)!=0 )
				g_print("Can't close file: '%s' (%s)\n",g_basename(mp3file->file),
					file!=NULL?g_strerror(errno):"file=NULL");
			return;
		}

		/* Read the header from file */
		fseek(file,position,SEEK_SET);
		fread(&buffer,1,MPEG_HEADER_LENGTH,file);
		/*g_print("\n%8x %8x %8x %8x",buffer[0],buffer[1],buffer[2],buffer[3]);*/

		/* Now read it to check if it's valid */
		switch ((buffer[1]>>3 & 0x03))
		{
			case 3:	/* MPEG 1 */
				mpegfile.version = 0;
				break;
			case 2:	/* MPEG 2 */
				mpegfile.version = 1;
				break;
			case 1: /* Reserved */
				break;
			case 0:	/* MPEG 2.5 */
				mpegfile.version = 2;
				break;
		}

		/* A protection against garbage which may give a invalid version */
		if ( mpegfile.version != 0 && mpegfile.version != 1 && mpegfile.version != 2 )
			goto Bad_Header;

		mpegfile.layer            = 3-((buffer[1]>>1) & 0x03);
		mpegfile.error_protection = !(buffer[1] & 0x01);
		mpegfile.bitrate          = (buffer[2]>>4) & 0x0F;
		mpegfile.bitrate          = bitrates[mpegfile.version][mpegfile.layer][mpegfile.bitrate-1];
		mpegfile.samplerate       = (buffer[2]>>2) & 0x03;
		mpegfile.samplerate       = frequencies[mpegfile.version][mpegfile.samplerate];
		mpegfile.padding          = (buffer[2]>>1) & 0x01;
		mpegfile.extension        = buffer[2] & 0x01;
		mpegfile.mode             = (buffer[3]>>6) & 0x03;
		mpegfile.mode_extention   = (buffer[3]>>4) & 0x03;
		mpegfile.copyright        = (buffer[3]>>3) & 0x01;
		mpegfile.original         = (buffer[3]>>2) & 0x01;
		mpegfile.emphasis         = (buffer[3]) & 0x03;

		if (mpegfile.samplerate == 0)
			goto Bad_Header;

		/* Calculate framezise */
		switch (mpegfile.layer)
		{
			case 1:
				mpegfile.framesize = (12*(mpegfile.bitrate*1000)/mpegfile.samplerate
						     +mpegfile.padding)*4;
				break;
			case 2:
			case 3:
				mpegfile.framesize = 144*mpegfile.bitrate*1000/mpegfile.samplerate
						     +mpegfile.padding;
				break;
		}

		/* Check if at the end of this frame, begins an other frame */
		if (Header_Find_Position(file,position+mpegfile.framesize) == position+mpegfile.framesize)
			break;

		/* Bad header: Try to find other header */
		Bad_Header:
			position++;
	}
	if ( !file || fclose(file)!=0 )
		g_print("Can't close file: '%s' (%s)\n",g_basename(mp3file->file),
							file!=NULL?g_strerror(errno):"file=NULL");

	/* Get size */
	stat(mp3file->file,&statbuf);
	filesize = (long)statbuf.st_size;

	/* Get the duration of the file */
	/* Note: Garbage at the start and tag aren't suppress from the file size, so duration
	 * may be a little superior than real duration */
	if (mpegfile.samplerate!=0 && mpegfile.framesize!=0)
	{
		mpegfile.totalframes = (long)((double)(filesize)/mpegfile.framesize);
		mpegfile.duration = mpegfile.totalframes*((float)1152/(float)(mpegfile.version+1))/
			(float)mpegfile.samplerate;
	}else
	{
		g_print("Warning!! framesize=%d samplerate=%d in file '%s', can't determine totalframes!\n",
			mpegfile.framesize,mpegfile.samplerate,g_basename(mp3file->file));
	}
	
/*	g_print("\ns %d %d %d %d\n",mpegfile.version,mpegfile.framesize,
				(gint)mpegfile.totalframes,mpegfile.duration);
*/
	/* Transfert needed values to MP3 struct */
	mp3file->mp3.version    = mpegfile.version;
	mp3file->mp3.layer      = mpegfile.layer;
	mp3file->mp3.bitrate    = mpegfile.bitrate;
	mp3file->mp3.samplerate = mpegfile.samplerate;
	mp3file->mp3.mode       = mpegfile.mode;
	mp3file->mp3.size       = filesize;
	mp3file->mp3.duration   = mpegfile.duration;

	TotalSize += filesize;
	TotalDuration += mpegfile.duration;
}



/*
 * Return the marker of the position of the header if succeed, else return -1 
 * if the end of file is reached (no header found)
 */
glong Header_Find_Position (FILE *file, long start_position)
{
	guchar buffer[MPEG_HEADER_LENGTH];
	size_t readsize;

	if (start_position == 0)
		rewind(file);
	else
		fseek(file,start_position,SEEK_SET);

	do
	{
		readsize = fread(&buffer,1,MPEG_HEADER_LENGTH,file);
		fseek(file,1-readsize,SEEK_CUR);
		if (readsize<4)
		{	
			/* End of file reached */
			//g_print("End of file reached!\n");
			//fclose(file);
			return -1;
		}
	}
	while ( ((buffer[0]&0xFF)!=0xFF) || ((buffer[1]&0xE0)!=0xE0) || ((buffer[2]&0xF0)==0xF0) );

	/* Return the starting position of header */
	return ftell(file)-1;
}




/*
 * Display header infos in the main window
 */
void Header_Display (MP3_File *mp3file)
{
	gchar *text;
	gchar *time = NULL;
	gchar *time1 = NULL;
	gchar *size = NULL;
	gchar *size1 = NULL;

	/* MPEG, Layer versions */
	text = g_strdup_printf("%s, Layer %s",version_names[mp3file->mp3.version],
					      layer_names[mp3file->mp3.layer]);
	gtk_label_set_text(GTK_LABEL(MpegLayerLabel),text);
	g_free(text);

	/* Bitrate */
	text = g_strdup_printf(_("%d kb/s"),mp3file->mp3.bitrate);
	gtk_label_set_text(GTK_LABEL(BitrateLabel),text);
	g_free(text);

	/* Samplerate */
	text = g_strdup_printf(_("%d Hz"),mp3file->mp3.samplerate);
	gtk_label_set_text(GTK_LABEL(SamplerateLabel),text);
	g_free(text);

	/* Mode */
	text = g_strdup_printf("%s",_(mode_names[mp3file->mp3.mode]));
	gtk_label_set_text(GTK_LABEL(ModeLabel),text);
	g_free(text);

	/* Size */
	size = Convert_Size(mp3file->mp3.size);
	size1 = Convert_Size(TotalSize);
	text = g_strdup_printf("%s (%s)",size,size1);
	gtk_label_set_text(GTK_LABEL(SizeLabel),text);
	g_free(size);
	g_free(size1);
	g_free(text);

	/* Duration */
	time = Convert_Duration(mp3file->mp3.duration);
	time1 = Convert_Duration(TotalDuration);
	text = g_strdup_printf("%s (%s)",time,time1);
	gtk_label_set_text(GTK_LABEL(DurationLabel),text);
	g_free(time);
	g_free(time1);
	g_free(text);
}


gchar *Convert_Size (gfloat size)
{
	gchar *data=NULL;
	/* Units Tab of file size (bytes,kilobytes,...) */
	gchar *Units_Tab[] = { N_("B"), N_("KB"), N_("MB"), N_("GB"), N_("TB")};
	gint i = 0;

	while ( (gint)size/1024 && i<(sizeof(Units_Tab)/sizeof(Units_Tab[0])-1) )
	{
		size = size/1024;
		i++;
	}
	return data = g_strdup_printf("%.1f %s",size,_(Units_Tab[i]));	
}


gchar *Convert_Duration (long int duration)
{
	guint hour=0;
	guint minute=0;
	guint second=0;
	gchar *data;

	if (duration<0) return "";

	hour = duration/3600;
	minute = (duration%3600)/60;
	second = (duration%3600)%60;
	/*g_print("\n%ld: %.2d %.2d %.2d\n",duration,hour,minute,second);*/

	if(hour)
		data = g_strdup_printf("%d:%.2d:%.2d",hour,minute,second);
	else
		data = g_strdup_printf("%d:%.2d",minute,second);
	
	return data;
}


/* 
 * Check if file contains MP3 header (is a mp3), returns 1 if good, else 0.
 */
/*int Check_If_Mp3_By_Header (gchar *filename)
 *{
 *	FILE *file;
 *
 *	if ( (file=fopen(filename,"r"))==0 )
 *	{
 *		g_print(_("ERROR while opening file: '%s' (%s)\n\a"),filename,g_strerror(errno));
 *		return 0;
 *	}
 *	if (Header_Find_Header(file)!=-1)
 *	{
 *		fclose(file);
 *		return 1;
 *	}
 *	fclose(file);
 *	return 0;
 *}
*/
